from __future__ import annotations

import asyncio
import aiohttp
import logging

from .utils.token import validate_token

from .api.api_client import ApiClient
from .api.manager import Manager
from .dispatcher import Dispatcher

from .api.api_references import ApiReference

from .models.models import Discussion

class Bot:
    def __init__(self, token: str, bot_id: int, api_reference: ApiReference, interval: int = 10, discussion_id: int = None, session: aiohttp.ClientSession | None = None) -> None:
        validate_token(token)

        self._token = token
        self._bot_id = bot_id
        self._discussion_id = discussion_id
        self._interval = interval

        self.api_reference = api_reference

        self.api_client = ApiClient(token, session=session, bot_id=bot_id)
        self.manager = Manager(self.api_client, discussion_id=discussion_id, bot_id=self._bot_id, api_reference=self.api_reference)

        self.dispatcher = Dispatcher()
        self._last_discussion_id: str | None = None

    def command(self, name: str):
        return self.dispatcher.command(name)

    def on_new_discussion(self, func):
        return self.dispatcher.on_new_discussion(func)

    def on_new_message(self, func):
        return self.dispatcher.on_new_message(func)

    async def start(self):
        logging.info("Starting bot...")
        async with ApiClient(self.token, self.bot_id) as client:
            self.api_client = client
            self.manager = Manager(client, self.discussion_id, self.bot_id, self.api_reference)

            await self._listen_loop()

    async def _listen_loop(self):
        logging.info("Listening for new posts and discussions...")
        while True:
            try:
                if self.discussion_id is not None:
                    new_posts = await self.manager.fetch_new_posts()
                    for post_id in new_posts:
                        msg = await self.manager.parse_post(post_id)
                        if not msg or not msg.message:
                            continue
                        logging.info(f"New message: {msg.message}")
                        await self.dispatcher.handle_message(msg, self)

                discussion: Discussion | None = await self.manager.fetch_new_discussion()
                if discussion:
                    logging.info(f"New discussion detected: {discussion.title}")
                    await self.dispatcher.handle_new_discussion(discussion)

                await asyncio.sleep(self.interval)

            except asyncio.TimeoutError:
                logging.warning("Timeout in listen loop (network issue?)")
            except Exception as e:
                logging.exception(f"Error in listen loop: {e}")

    def run(self):
        asyncio.run(self.start())

    @property
    def token(self) -> str:
        return self._token

    @property
    def bot_id(self) -> int:
        return self._bot_id

    @property
    def discussion_id(self) -> int:
        return self._discussion_id

    @property
    def interval(self) -> int:
        return self._interval

    @property
    def session(self) -> aiohttp.ClientSession:
        return self.session

    async def close(self) -> None:
        if self.api_client and self.api_client.session and not self.api_client.session.closed:
            await self.api_client.session.close()


from dataclasses import dataclass
from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from ..api.manager import Manager

class NotAttachedToManager(Exception):
    pass

@dataclass
class Message:
    post_id: str
    message: str
    reply_id: Optional[str]
    user_id: Optional[str]
    username: Optional[str]
    _manager: Optional["Manager"] = None

    async def edit(self, content: str):
        if not self._manager:
            raise NotAttachedToManager("Message is not attached to a manager")
        return await self._manager.edit_post(content, int(self.post_id))

    async def delete(self):
        if not self._manager:
            raise NotAttachedToManager("Message is not attached to a manager")
        return await self._manager.delete_post(int(self.post_id))

    async def like(self):
        if not self._manager:
            raise NotAttachedToManager("Message is not attached to a manager")
        return await self._manager.like_post(int(self.post_id))

    async def answer(self, content:str):
        if not self._manager:
            raise NotAttachedToManager("Message is not attached to a manager")
        return await self._manager.create_post(content)

    async def reply(self, content: str):
        if not self._manager:
            raise NotAttachedToManager("Message is not attached to a manager")
        return await self._manager.create_post(content, reply=True, reply_to=self)

    async def parse_user(self):
        if not self._manager:
            raise NotAttachedToManager("Message is not attached to a manager")
        return await self._manager.parse_user(int(self.user_id))


@dataclass
class User:
    username: str
    display_name: str
    slug: str
    joined_at: Optional[str]
    discussions_count: int
    comments_count: int
    last_seen_at: Optional[str]
    steam_id: str
    bio: str
    rank: str


@dataclass
class Discussion:
    id: str
    title: str
    slug: str
    comments_count: int
    participants_count: int
    created_at: Optional[str]
    updated_at: Optional[str]
    content: str
    tags: list[str]
    first_post_id: str
    raw: dict
    _manager: Optional["Manager"] = None

    async def reply(self, content: str):
        if not self._manager:
            raise NotAttachedToManager("Discussion is not attached to a manager")
        old_discussion_id = self._manager.discussion_id
        self._manager.discussion_id = int(self.id)

        try:
            return await self._manager.create_post(content)
        finally:
            self._manager.discussion_id = old_discussion_id

    async def edit_first_post(self, content: str):
        if not self._manager:
            raise NotAttachedToManager("Discussion is not attached to a manager")
        return await self._manager.edit_post(content, int(self.first_post_id))

    async def delete(self):
        if not self._manager:
            raise NotAttachedToManager("Discussion is not attached to a manager")
        return await self._manager.delete_discussion(int(self.id))

    async def lock(self):
        if not self._manager:
            raise NotAttachedToManager("Discussion is not attached to a manager")
        return await self._manager.lock_discussion(int(self.id))

    async def approve(self):
        if not self._manager:
            raise NotAttachedToManager("Discussion is not attached to a manager")
        return await self._manager.add_tag(int(self.id), 9)

    async def reject(self):
        if not self._manager:
            raise NotAttachedToManager("Discussion is not attached to a manager")
        return await self._manager.add_tag(int(self.id), 13)



class ApiReference:
    def __init__(self, base_url: str):
        self.base_url = base_url.rstrip("/")

    def url(self, path: str) -> str:
        return f"{self.base_url}/{path.lstrip('/')}"

    @property
    def users(self):
        return self.url("users")

    def user(self, user_id: int):
        return self.url(f"users/{user_id}")

    @property
    def posts(self):
        return self.url("posts")

    def post(self, post_id: int):
        return self.url(f"posts/{post_id}")

    @property
    def discussions(self):
        return self.url("discussions")

    @property
    def discussions_last(self):
        return self.url("discussions?include=user,lastPostedUser,tags,tags.parent,recipientUsers,recipientGroups,firstPost&sort=-createdAt&page[offset]=0")

    def discussion(self, discussion_id: int):
        return self.url(f"discussions/{discussion_id}")



import asyncio
import json
import logging
from typing import Optional

from ..utils.html_stripper import HTMLStripper

from ..models.models import Message, User, Discussion

from .api_references import ApiReference

HTMLStripper = HTMLStripper()
KNOWN_POSTS_FILE = "known_posts.json"
KNOWN_DISCUSSIONS_FILE = "known_discussions.json"


class Manager:
    def __init__(self, api_client, discussion_id: int, bot_id: int, api_reference: ApiReference):
        self.api_client = api_client
        self.discussion_id = discussion_id
        self.bot_id = bot_id
        self.api_reference = api_reference
        self.known_posts: set = self._load_known_posts()
        self.known_discussions: set = self._load_known_discussions()

    def _load_known_posts(self) -> set:
        try:
            with open(KNOWN_POSTS_FILE, "r") as file:
                data = json.load(file)
                return set(data)
        except (FileNotFoundError, json.JSONDecodeError):
            logging.warning("No known posts file found, starting fresh.")
            empty = set()
            with open(KNOWN_POSTS_FILE, "w") as file:
                json.dump([], file)
            return empty

    def _save_known_posts(self):
        with open(KNOWN_POSTS_FILE, "w") as file:
            json.dump(list(self.known_posts), file)

    def _load_known_discussions(self) -> set:
        try:
            with open(KNOWN_DISCUSSIONS_FILE, "r") as file:
                data = json.load(file)
                return set(data)
        except (FileNotFoundError, json.JSONDecodeError):
            logging.warning("No known discussions file found, starting fresh.")
            empty = set()
            with open(KNOWN_DISCUSSIONS_FILE, "w") as file:
                json.dump([], file)
            return empty

    def _save_known_discussions(self):
        with open(KNOWN_DISCUSSIONS_FILE, "w") as file:
            json.dump(list(self.known_discussions), file)

    async def create_post(self, content: str, reply: bool = False, reply_to: Optional[Message] = None, discussion_id: int = None) -> Message:
        if reply and reply_to:
            username = reply_to.username
            post_id = reply_to.post_id
            content = f'@"{username}"#p{post_id} {content}'

        url = self.api_reference.posts
        data = {
            "data": {
                "type": "posts",
                "attributes": {"content": content},
                "relationships": {
                    "discussion": {
                        "data": {"type": "discussions", "id": discussion_id or self.discussion_id}
                    }
                },
            }
        }

        result = await self.api_client.request("POST", url, data=data)

        if result.get("status") != 201:
            return result

        logging.info("Successfully created a new post.")

        raw = result["json"]
        post_data = raw.get("data", {})
        included = raw.get("included", [])

        content_html = post_data.get("attributes", {}).get("contentHtml", "")
        message, reply_id = HTMLStripper.strip_html_and_extract_reply_id(content_html)

        user_id = post_data.get("relationships", {}).get("user", {}).get("data", {}).get("id")
        username = None
        for item in included:
            if item.get("type") == "users" and item.get("id") == user_id:
                username = item.get("attributes", {}).get("username")
                break

        return Message(
            post_id=post_data.get("id"),
            message=message,
            reply_id=reply_id,
            user_id=user_id,
            username=username,
            _manager=self
        )

    async def edit_post(self, content: str, post_id: int) -> dict:
        url = self.api_reference.post(post_id)
        data = {
            "data": {
                "type": "posts",
                "id": post_id,
                "attributes": {
                    "content": content
                }
            }
        }
        return await self.api_client.request("PATCH", url, data=data)

    async def edit_bio(self, content: str) -> dict:
        url = self.api_reference.user(self.bot_id)
        data = {
            "data": {
                "type": "users",
                "id": self.bot_id,
                "attributes": {"bio": content}
            }
        }
        return await self.api_client.request("PATCH", url, data=data)

    async def like_post(self, post_id: int) -> dict:
        url = self.api_reference.post(post_id)
        data = {
            "data": {
                "type": "posts",
                "attributes": {
                    "isLiked": True
                },
                "id": post_id
            }
        }
        return await self.api_client.request("PATCH", url, data=data)

    async def delete_post(self, post_id: int) -> dict:
        url = self.api_reference.post(post_id)
        data = {
            "data": {
                "type": "posts",
                "id": str(post_id),
                "attributes": {
                    "isHidden": True
                }
            }
        }
        return await self.api_client.request("PATCH", url, data=data)

    async def delete_discussion(self, discussion_id: int) -> dict:
        url = self.api_reference.discussion(discussion_id)
        data = {
            "data": {
                "type": "discussions",
                "id": str(discussion_id),
                "attributes": {
                    "isHidden": True
                }
            }
        }
        return await self.api_client.request("PATCH", url, data=data)

    async def lock_discussion(self, discussion_id: int) -> dict:
        url = self.api_reference.discussion(discussion_id)
        data = {
            "data": {
                "type": "discussions",
                "id": str(discussion_id),
                "attributes": {
                    "isLocked": True
                }
            }
        }
        return await self.api_client.request("PATCH", url, data=data)

    async def add_tag(self, discussion_id: int, tag_id: int) -> dict:
            url = self.api_reference.discussion(discussion_id)
            discussion = await self.parse_discussion(discussion_id)
            
            if not discussion:
                return {"error": "Discussion not found"}

            unique_tags = set(discussion.tags)
            unique_tags.add(str(tag_id))
            
            data = {
                "data": {
                    "type": "discussions",
                    "id": str(discussion_id),
                    "attributes": {},
                    "relationships": {
                        "tags": {
                            "data": [{"type": "tags", "id": tid} for tid in unique_tags]
                        }
                    }
                }
            }
            
            return await self.api_client.request("PATCH", url, data=data)

    async def fetch_new_posts(self, max_retries: int = 3) -> list[str]:
        url = self.api_reference.discussion(self.discussion_id)

        for attempt in range(1, max_retries + 1):
            result = await self.api_client.request("GET", url)

            if result.get("status") != 200:
                logging.warning(f"Attempt {attempt}: Failed to fetch discussion")
                await asyncio.sleep(2)
                continue

            posts = (
                result.get("json", {})
                .get("data", {})
                .get("relationships", {})
                .get("posts", {})
                .get("data", [])
            )
            new_posts = [p["id"] for p in posts if p["id"] not in self.known_posts]

            if new_posts:
                self.known_posts.update(new_posts)
                self._save_known_posts()
                return new_posts

            return []

        return []

    async def parse_post(self, post_id: str) -> Optional[Message]:
        url = self.api_reference.post(int(post_id))
        result = await self.api_client.request("GET", url)
        data = result.get("json", {})
        post_data = data.get("data", {})

        message, reply_id = HTMLStripper.strip_html_and_extract_reply_id(post_data.get("attributes", {}).get("contentHtml", ""))

        username = None
        for item in data.get("included", []):
            if item.get("type") == "users":
                username = item.get("attributes", {}).get("username")
                break

        return Message(
            post_id=post_id,
            message=message,
            reply_id=reply_id,
            user_id=post_data.get("relationships", {}).get("user", {}).get("data", {}).get("id"),
            username=username,
            _manager=self
        )

    async def parse_user(self, user_id: int) -> Optional[User]:
        url = self.api_reference.user(user_id)
        result = await self.api_client.request("GET", url)

        data = result.get("json", {})
        attributes = data.get("data", {}).get("attributes", {})

        return User(
            username=attributes.get("username"),
            display_name=attributes.get("displayName"),
            slug=attributes.get("slug"),
            joined_at=attributes.get("joinTime"),
            discussions_count=attributes.get("discussionCount"),
            comments_count=attributes.get("commentCount"),
            last_seen_at=attributes.get("lastSeenAt") or "Последний заход приватный",
            steam_id=attributes.get("SteamAuth", {}).get("identifier"),
            bio=attributes.get("bio"),
            rank = (data.get("included") or [{}])[0].get("attributes", {}).get("nameSingular", "Ранга нет")
        )

    async def parse_discussion(self, discussion_id: int) -> Optional[Discussion]:
        url = self.api_reference.discussion(discussion_id)
        result = await self.api_client.request("GET", url)

        data = result.get("json", {})
        discussion = data.get("data", {})
        attributes = discussion.get("attributes", {})
        relationships = discussion.get("relationships", {})
        included = data.get("included", [])

        content, reply_id = None, None
        if included:
            first_post = next((i for i in included if i.get("type") == "posts"), None)
            if first_post:
                raw_html = first_post.get("attributes", {}).get("contentHtml", "")
                content, reply_id = HTMLStripper.strip_html_and_extract_reply_id(raw_html)

        try:
            tags = [t["id"] for t in relationships.get("tags", {}).get("data", [])]
        except: #type: ignore
            tags = None
        return Discussion(
            id=discussion.get("id"),
            title=attributes.get("title"),
            slug=attributes.get("slug"),
            comments_count=attributes.get("commentCount"),
            participants_count=attributes.get("participantCount"),
            created_at=attributes.get("createdAt"),
            updated_at=attributes.get("lastPostedAt"),
            content=content,
            tags=tags,
            first_post_id=included[0]["id"] if included else None,
            raw=data,
            _manager=self
        )

    async def get_last_discussion(self) -> Optional[Discussion]:
        url = self.api_reference.discussions_last
        result = await self.api_client.request("GET", url)
        data = result.get("json", {})
        discussion = data["data"][0]

        return await self.parse_discussion(discussion["id"])

    async def fetch_new_discussion(self) -> Optional[Discussion]:
        discussion = await self.get_last_discussion()
        if not discussion:
            return None

        if discussion.id in self.known_discussions:
            return None

        self.known_discussions.add(discussion.id)
        self._save_known_discussions()
        return discussion




class RequestFormatter:
    @staticmethod
    def format_status(status: int) -> str:
        return {
            200: "Success",
            201: "Created",
            400: "Bad request",
            401: "Unauthorized",
            403: "Forbidden",
            404: "Not found",
            405: "Method not allowed",
            500: "Internal server error",
        }.get(status, f"Unknown error (code {status})")


from functools import lru_cache

class TokenValidationError(Exception):
    pass

@lru_cache()
def validate_token(token: str) -> bool:
    if not isinstance(token, str):
        raise TokenValidationError(
            f"Token type {type(token)} not supported, must be str."
        )

    if any(x.isspace() for x in token):
        raise TokenValidationError(
            f"Token can't contain spaces."
        )

    return True


from html.parser import HTMLParser

class HTMLStripper(HTMLParser):
    def __init__(self):
        super().__init__()
        self.result = []
        self.reply_post_id = None

    def handle_starttag(self, tag, attrs):
        if tag == "a":
            attr_dict = dict(attrs)
            if attr_dict.get("class") == "PostMention" and "data-id" in attr_dict:
                self.reply_post_id = attr_dict["data-id"]

    def handle_data(self, d):
        self.result.append(d)

    def get_data(self):
        return ''.join(self.result).strip()

    def get_reply_id(self):
        return self.reply_post_id

    @staticmethod
    def strip_html_and_extract_reply_id(html: str) -> tuple[str, str | None]:
        stripper = HTMLStripper()
        stripper.feed(html)
        return stripper.get_data(), stripper.get_reply_id()


from __future__ import annotations

import logging
import inspect
from typing import Callable, Awaitable, Any

class Dispatcher:
    def __init__(self):
        self._commands: dict[str, Callable[..., Awaitable[Any]]] = {}
        self._new_discussion_handlers: list[Callable[..., Awaitable[Any]]] = []
        self._new_message_handlers: list[Callable[..., Awaitable[Any]]] = []

    def command(self, name: str):
        def decorator(func: Callable[..., Awaitable[Any]]):
            self._commands[name] = func
            return func
        return decorator

    def on_new_discussion(self, func: Callable[..., Awaitable[Any]]):
        self._new_discussion_handlers.append(func)
        return func

    def on_new_message(self, func: Callable[..., Awaitable[Any]]):
        self._new_message_handlers.append(func)
        return func

    async def _call_handler(self, handler: Callable[..., Any], msg, bot):
        try:
            sig = inspect.signature(handler)
            params_count = len(sig.parameters)
        except Exception:
            params_count = 1

        if params_count <= 1:
            call_args = (msg,)
        else:
            call_args = (msg, bot)

        try:
            result = handler(*call_args)
            if inspect.isawaitable(result):
                await result
        except Exception:
            logging.exception("Error while running handler %s", getattr(handler, "__name__", repr(handler)))

    async def handle_message(self, msg, bot):
        if int(msg.user_id) == int(bot.bot_id):
            logging.info("Skipping my own message..")
            return

        logging.info("Handling message")

        parts = (msg.message or "").split()
        if not parts:
            return
        cmd = parts[0]
        if cmd in self._commands:
            await self._call_handler(self._commands[cmd], msg, bot)
            return

        for handler in list(self._new_message_handlers):
            await self._call_handler(handler, msg, bot)

    async def handle_new_discussion(self, discussion):
        logging.info(f"Handling new discussion: {getattr(discussion, 'id', '?')}")
        for handler in list(self._new_discussion_handlers):
            await self._call_handler(handler, discussion, None)



import aiohttp
import logging
import asyncio
from typing import Optional

from ..utils.formatter import RequestFormatter

class ApiClient:
    def __init__(self, token: str, bot_id: int, max_retries: int = 5, session: aiohttp.ClientSession | None = None):
        self._token = token
        self._bot_id = bot_id
        self._session: aiohttp.ClientSession | None = session
        self.max_retries = max_retries

    async def __aenter__(self):
        self._session = aiohttp.ClientSession(
            headers={
                "Authorization": f"Token {self._token}; userId={self._bot_id}",
                "Content-Type": "application/json",
                "Accept": "application/json",
            }
        )
        return self

    async def __aexit__(self, exc_type, exc, tb):
        if self._session and not self._session.closed:
            await self._session.close()

    @property
    def session(self) -> aiohttp.ClientSession:
        if self._session is None or self._session.closed:
            self._session = aiohttp.ClientSession(
                headers={
                    "Authorization": f"Token {self._token}; userId={self._bot_id}",
                    "Content-Type": "application/json",
                }
            )
        return self._session

    async def request(self, method: str, url: str, data: Optional[dict] = None, timeout: int = 10) -> Optional[dict]:
        for attempt in range(1, self.max_retries + 1):
            try:
                async with self.session.request(method, url, json=data, timeout=aiohttp.ClientTimeout(total=timeout)) as response:
                    if response.status == 429:
                        delay = 10
                        logging.warning(f"Too many requests. Retry in {delay} seconds.")
                        await asyncio.sleep(delay)
                        continue

                    result = {
                        "status": response.status,
                        "status_text": RequestFormatter.format_status(response.status),
                        "json": await response.json(),
                    }
                    logging.debug(f"{method} {url} -> {response.status}")
                    return result

            except aiohttp.ClientError as e:
                logging.error(f"Request error: {e}")
                await asyncio.sleep(2**attempt)
                return {"error": str(e)}

    async def close(self) -> None:
        if self._session and not self._session.closed:
            await self._session.close()